from ..utilities.string_converters import *
import numpy as np

class MetadataMetric:
    """A metric for comparing metadata.

    This class is designed to be used as a callable object that takes
    two collections of metadata (`sxs.Metadata`, `dict`, `pd.Series`)
    and returns a number measuring the distance between the metadata.
    
    This is intended to be used as a heuristic for sorting and
    filtering metadata, rather than as a strict metric for clustering
    or classification.  In particular, note that this does not account
    for the fact that parameters are typically measured at the
    reference time, which can be quite different for different
    systems.

    Note that calling an object of this class with two metadata
    collections will return the *squared* distance between them.

    Parameters
    ----------
    parameters : list of str, optional
        The names of the metadata fields to be compared.  The defaults
        are the mass ratio, spins, and eccentricity at the reference
        time.  Note that all of these fields *must* be present in
        *both* metadata collections — except for
        `reference_mass_ratio` or `reference_complex_eccentricity`
        which will be calculated from other parameters if not present.
        (The `Metadata.add_standard_parameters` method may be useful
        here.)
    metric : array_like, optional
        The matrix used to weight the differences in the parameters.
        The default is a diagonal matrix with ones on the diagonal.
    allow_different_object_types : bool, optional
        If True, metadata with different object types (BHBH, BHNS,
        NSNS) will be compared without penalty.  If False, metadata
        with different object types will be assigned an infinite
        distance.
    eccentricity_threshold1 : float, optional
        The threshold eccentricity below which we consider metadata1
        non-eccentric.  Default is 1e-2.
    eccentricity_threshold2 : float, optional
        The threshold eccentricity below which we consider metadata2
        non-eccentric.  Default is 1e-3.
    eccentricity_threshold_penalize_shorter : int, optional
        The number of orbits below which we penalize metadata2 for
        having a non-zero eccentricity when metadata1 does not.  This
        is intended to avoid ascribing small distances to systems with
        shorter inspirals.  Default is 20.

    The eccentricity is ignored if eccentricity is one of the
    parameters to be compared and all three of the following are true:
      1) the eccentricity in metadata1 is below
         `eccentricity_threshold1`,
      2) the eccentricity in metadata2 is below
         `eccentricity_threshold2`, and
      3) the number of orbits in metadata2 is longer than
         `eccentricity_threshold_penalize_shorter`.
    You may set these arguments to 0 to disable these features.

    """
    def __init__(
            self,
            parameters=[
                "reference_mass_ratio",
                "reference_dimensionless_spin1",
                "reference_dimensionless_spin2",
                "reference_complex_eccentricity",
            ],
            metric=None,
            allow_different_object_types=False,
            eccentricity_threshold1=1e-2,
            eccentricity_threshold2=1e-3,
            eccentricity_threshold_penalize_shorter=20,
    ):
        self.parameters = parameters
        self.metric = metric
        self.allow_different_object_types = allow_different_object_types
        self.eccentricity_threshold1 = eccentricity_threshold1
        self.eccentricity_threshold2 = eccentricity_threshold2
        self.eccentricity_threshold_penalize_shorter = eccentricity_threshold_penalize_shorter

    def __call__(self, metadata1, metadata2, debug=False):
        if not self.allow_different_object_types:
            type1 = (
                metadata1["object_types"]
                if "object_types" in metadata1
                else "".join(sorted([
                    metadata1.get("object1", "A").upper(),
                    metadata1.get("object2", "B").upper()
                ]))
            )
            type2 = (
                metadata2["object_types"]
                if "object_types" in metadata2
                else "".join(sorted([
                    metadata2.get("object1", "C").upper(),
                    metadata2.get("object2", "D").upper()
                ]))
            )
            if type1 != type2:
                return np.inf
        
        # Create empty lists because the element type will vary;
        # each element could be `float`, `complex`, or `list`.
        values1 = [np.nan] * len(self.parameters)
        values2 = [np.nan] * len(self.parameters)

        # Fill in the values with the metadata
        for values, metadata in [(values1, metadata1), (values2, metadata2)]:
            for i, parameter in enumerate(self.parameters):
                if parameter in metadata:
                    values[i] = metadata[parameter]
                elif parameter == "reference_mass_ratio":
                    values[i] = (
                        floater(metadata.get("reference_mass1", np.nan))
                        / floater(metadata.get("reference_mass2", np.nan))
                    )
                elif parameter == "reference_complex_eccentricity":
                    e = floaterbound(
                        metadata.get(
                            "reference_eccentricity_bound",
                             metadata.get("reference_eccentricity", np.nan)
                        )
                    )
                    l = floater(metadata.get("reference_mean_anomaly", np.nan))
                    values[i] = e * np.exp(1j * l)

        if debug:
            print(f"{self.parameters=}")
            print(f"{values1=}")
            print(f"{values2=}")

        if "reference_eccentricity" in self.parameters or "reference_complex_eccentricity" in self.parameters:
            i = (
                self.parameters.index("reference_eccentricity")
                if "reference_eccentricity" in self.parameters
                else self.parameters.index("reference_complex_eccentricity")
            )

            if abs(values1[i]) < self.eccentricity_threshold1:
                # Then we consider metadata1 a non-eccentric system...

                # ...and we ignore the eccentricity if metadata2 is also non-eccentric,
                # and longer than eccentricity_threshold_penalize_shorter.
                if (
                    abs(values2[i]) < self.eccentricity_threshold2
                    and metadata2.get(
                        "number_of_orbits",
                        metadata2.get("number_of_orbits_from_start", 0)
                    ) > self.eccentricity_threshold_penalize_shorter
                ):
                    values1[i] = values2[i]

        # Concatenate the values into a single 1-d array (even if some
        # entries are 3-vectors).
        difference = (
            np.concatenate(list(map(np.atleast_1d, values1)))
            - np.concatenate(list(map(np.atleast_1d, values2)))
        )

        if debug:
            print(f"{difference=}")

        metric = self.metric or np.diag(np.ones(len(difference)))

        return np.real(difference @ metric @ difference.conj())



#!/usr/bin/env python3


def fetch_RIT_catalog_data():
    """Fetch metadata from the RIT Catalog and parse as a dict.

    Scrape the RIT Catalog page and extract information about the numerical
    simulations. The RIT Catalog can be accessed through
    https://ccrg.rit.edu/numerical-simulations, but the actual data is hosted at
    https://ccrgpages.rit.edu/~RITCatalog/. Retrieve the simulation IDs,
    download links for extrapolated Psi4 and Strain data, and the metadata for
    each simulation. The metadata is parsed into a dict.

    Returns
    =======
    dict
        A dictionary mapping simulation name to metadata.

    """

    from ...metadata.metadata import Metadata

    import requests
    import tempfile
    from bs4 import BeautifulSoup

    RITcatalog_url = "https://ccrgpages.rit.edu/~RITCatalog/"

    session = requests.Session()

    response = session.get(RITcatalog_url)

    # Check if the request was successful
    if response.status_code != 200:
        raise Exception(f"Failed to load page {RITcatalog_url}")

    soup = BeautifulSoup(response.text, "html.parser")

    # Finds the table with id "example", i.e.,
    # the one that has the numerical simulations data

    table = soup.find("table", id="example")

    # Goes to tbody (table body) and finds all rows (tr)
    rows = table.find("tbody").find_all("tr")

    RITcatalogdata = {}

    # Goes through each row and extracts the data
    for row in rows:
        cells = row.find_all("td")
        if len(cells) == 0:
            continue

        sim_id = cells[0].get_text(strip=True)
        links = cells[2].find_all("a", href=True)

        if len(links) != 3:
            raise Exception(
                "Expecting 3 links, but getting an unexpected number of links "
                f"for simulation {sim_id}."
            )

        # Validate links contain expected keywords
        psi_link = strain_link = metadata_link = None
        for link in links:
            href = link["href"].lower()
            if "psi" in href:
                psi_link = link
            elif "strain" in href:
                strain_link = link
            elif "metadata" in href:
                metadata_link = link

        if not all([psi_link, strain_link, metadata_link]):
            raise Exception(f"Missing expected links for simulation {sim_id}.")

        RITcatalogdata[sim_id] = {}

        # Fetching metadata file
        metadata_response = session.get(RITcatalog_url + metadata_link["href"])
        if metadata_response.status_code != 200:
            raise Exception(f"Failed to load page for simulation {str(sim_id)}")

        # Parsing metadata content
        # Since from_txt_file needs a file path,
        # we create a temporary file

        with tempfile.NamedTemporaryFile(mode="w", suffix=".txt", delete=False) as tmp:
            tmp.write(metadata_response.text)
            tmp_path = tmp.name
            tmp.seek(0)
            RITcatalogdata[sim_id] = Metadata.from_txt_file(tmp_path, cache_json=False)

            # from_txt_file() adds metadata_path, which is meaningless here.
            # _backwards_compatibility() adds number_of_orbits=NaN, which we
            # don't want. Throw away both of these keys.
            RITcatalogdata[sim_id].pop("metadata_path", None)

        RITcatalogdata[sim_id]["extrap_psi4_url"] = RITcatalog_url + psi_link["href"]
        RITcatalogdata[sim_id]["extrap_strain_url"] = (
            RITcatalog_url + strain_link["href"]
        )
        RITcatalogdata[sim_id]["metadata_url"] = RITcatalog_url + metadata_link["href"]

    return RITcatalogdata


if __name__ == "__main__":

    import json
    import argparse

    # absolute imports in a standalone script
    from sxscatalog.utilities.convert_other_catalogs.convert_RIT_site import (
        fetch_RIT_catalog_data,
    )
    from sxscatalog.utilities import consolidate_xyz_vectors

    parser = argparse.ArgumentParser(
        description="Fetch RIT Catalog data and save to JSON."
    )

    parser.add_argument(
        "--output",
        type=argparse.FileType("w"),
        default="RIT_catalog.json",
        help="Output JSON file name (default: %(default)s).",
    )
    parser.add_argument(
        "--consolidate-xyz-vectors",
        action=argparse.BooleanOptionalAction,
        default=True,
        help="if triples of <key>x <key>y <key>z should be consolidated "
        "into vectors (default: %(default)s)",
    )

    args = parser.parse_args()

    data_dict = fetch_RIT_catalog_data()
    if args.consolidate_xyz_vectors:
        data_dict = {k: consolidate_xyz_vectors(v) for k, v in data_dict.items()}

    json.dump(data_dict, args.output, indent=2, sort_keys=True)



from .downloads import download_file
from .sxs_directories import (
    sxs_directory, read_config, write_config
)
from .sxs_identifiers import (
    sxs_identifier_regex, sxs_identifier_re,
    lev_regex, lev_re, lev_path_re,
    sxs_id_version_lev_regex, sxs_id_version_lev_re,
    sxs_id_version_lev_exact_regex, sxs_id_version_lev_exact_re,
    sxs_path_regex, sxs_path_re,
    sxs_id, sxs_id_and_version,
    lev_number, simulation_title, sxs_id_to_url,
)
from .files import (
    md5checksum, lock_file_manager, find_simulation_directories, find_files
)

def path_to_invenio(file_path):
    """Convert a file path to an invenio-compatible name"""
    import os
    return str(file_path).replace(os.path.sep, ":")


def invenio_to_path(file_name):
    """Convert an invenio-compatible name to a file path"""
    import os
    from pathlib import Path
    return Path(file_name.replace(":", os.path.sep))

def consolidate_xyz_vectors(data):
    """Find instances of key triples <key>x, <key>y, <key>z and consolidate into
    vectors named <key>.  If <key> ends with a dash (-) or underscore (_), the
    trailing dash/underscore will be stripped. The argument to this function is
    modified and returned.

    """

    names = []

    for k in data:
        if k[-1] == 'z':
            base = k[:-1]
            if (((base + 'x') in data) and
                ((base + 'y') in data)):
                names.append(base)

    for name in names:
        vec = [data.pop(name + 'x'),
               data.pop(name + 'y'),
               data.pop(name + 'z')]

        if name[-1] in ['-', '_']:
            name = name[:-1]

        data[name] = vec

    return data



#!/usr/bin/env python

if __name__ == "__main__":
    import argparse
    import json
    import pandas as pd
    from pathlib import Path

    from sxscatalog.utilities import consolidate_xyz_vectors

    parser = argparse.ArgumentParser(
        description="Compute the MAYA_catalog.json file from MAYAmetadata.pkl",
        epilog=(
            "Read the pandas dataframe of the pickle file, and "
            "arranges it so that it is similar to SXS catalog.json."
        ),
    )
    parser.add_argument(
        "filename",
        type=argparse.FileType("rb"),
        help="path to the MAYAmetadata.pkl file",
    )
    parser.add_argument(
        "--output",
        type=argparse.FileType("w"),
        default="MAYA_catalog.json",
        help="path to the output file (default: %(default)s).",
    )
    parser.add_argument(
        "--consolidate-xyz-vectors",
        action=argparse.BooleanOptionalAction,
        default=True,
        help="if triples of <key>x <key>y <key>z should be consolidated "
        "into vectors (default: %(default)s)",
    )

    args = parser.parse_args()

    data = pd.read_pickle(args.filename)
    data_dict = data.to_dict(orient="index")

    if args.consolidate_xyz_vectors:
        data_dict = {k: consolidate_xyz_vectors(v) for k, v in data_dict.items()}

    json.dump(data_dict, args.output, indent=2, sort_keys=True)



"""Interface to SXS metadata files"""

from .metadata import Metadata
from .metric import MetadataMetric

formats = {
    None: Metadata,
    "": Metadata,
    "metadata": Metadata,
}


def metadata_version(metadata):
    """Guess the version from the metadata keys"""

    # Define the keys that are present in each version
    keys_v1 = {
        "relaxed_mass1",
    }

    keys_v2 = {
        "metadata_version",
        "number_of_orbits",
    }

    keys_v3 = {
        "internal_changelog",
        "internal_minor_version",
        "metadata_content_revision",
        "metadata_format_revision",
        "number_of_orbits_from_reference_time",
        "number_of_orbits_from_start",
        #"t_relaxed_algorithm",
    }

    # Get the keys from the simulation metadata
    metadata_keys = set(metadata.keys())

    # Check for the presence of keys to determine the version
    if keys_v3.issubset(metadata_keys):
        return "v3.0"
    elif keys_v2.issubset(metadata_keys):
        return "v2.0"
    elif keys_v1.issubset(metadata_keys):
        return "v1.1"
    else:
        return ""
